from pathlib import Path
from dataclasses import dataclass
from aiohttp import ClientSession
import tenacity

@dataclass
class Provider:
    
    ID: str | int
    fileID: str | int
    url: str

    slug: str
    author: str

    def to_dict(self):
        
        data = {
            "ID": self.ID,
            "fileID": self.fileID,
            "url": self.url,
            "slug": self.slug,
            "author": self.author
        }

        return data

@dataclass
class Side:
    
    client: str
    server: str
    summary: str

    def to_dict(self):
        
        data = {
            "client": self.client,
            "server": self.server,
            "summary": self.summary,
        }

        return data

@dataclass
class Resource:

    "Represents downloadable item i.e. mod, resourcepack, shaderpack etc."

    name: str
    filename: str

    side: dict[str, dict[str]]
    hashes: dict[str, str | int]
    downloads: dict[str, dict[str]]

class ResourceAPI(object):

    def __init__(self, session: ClientSession) -> None:

        self.session = session

        self.session.headers["User-Agent"] = "Mozilla/5.0 (Macintosh; Intel Mac OS X x.y; rv:42.0) Gecko/20100101 Firefox/42.0"
        self.session.headers["Accept"] = "application/json"
        self.session.headers["Content-Type"] = "application/json"

        self.modrinth = "https://api.modrinth.com/v2"
        self.curseforge = "https://addons-ecs.forgesvc.net/api/v2"

        super().__init__()
    
    @tenacity.retry(stop=tenacity.stop_after_attempt(5), wait=tenacity.wait.wait_fixed(1))
    async def get(self, path: Path) -> Resource | None:

        from .utils import get_hash

        murmur2_hash = get_hash(path, "murmur2")
        sha512_hash = get_hash(path, "sha512")
        sha1_hash = get_hash(path, "sha1")

        modrinth_links = [
            f"{self.modrinth}/version_file/{sha512_hash}?algorithm=sha512",
            f"{self.modrinth}/version_file/{sha1_hash}?algorithm=sha1"
        ]
        
        modrinth = None
        for url in modrinth_links:
            async with self.session.get(url) as response:
                if response.status == 200 or response.status == 504:
                    json = await response.json()
                    modrinth = await self._get_modrinth(json)
                    break

        curseforge = None
        async with self.session.post(f"{self.curseforge}/fingerprint", data = f"[{murmur2_hash}]") as response:
            json = await response.json()
            if json['exactMatches']:
                curseforge = await self._get_curseforge(json['exactMatches'][0], murmur2_hash)

        if modrinth: 
            resource = modrinth
            if curseforge: 
                resource.downloads.update(curseforge.downloads)
                resource.hashes.update(curseforge.hashes)

        elif curseforge: resource = curseforge
        else: return

        return resource

    async def get_by_hash(self, hash_type: str, hash: str | int) -> Resource | None:

        match hash_type:

            case "sha1" | "sha256" | "sha512":

                url = f"{self.modrinth}/version_file/{hash}?algorithm={hash_type}"
                async with self.session.get(url) as response:
                    if response.status == 200:
                        json = await response.json()
                        return await self._get_modrinth(json)

            case "murmur2": 

                url = f"{self.curseforge}/fingerprint"              
                async with self.session.post(url, data = f"[{hash}]") as response:
                    json = await response.json()
                    if json['exactMatches']:
                        return await self._get_curseforge(json['exactMatches'][0], hash)

        return None

    async def get_by_ID(self, ID: str, fileID: str) -> Resource | None:

        url = f"{self.modrinth}/version_file/{fileID}"
        async with self.session.get(url) as response:
            if response.status == 200:
                json = await response.json()
                return await self._get_modrinth(json)

            else:
                url = f"{self.curseforge}/addon/{ID}/file/{fileID}"
                async with self.session.get(url) as response:
                    json = await response.json()
                    url = f"{self.curseforge}/fingerprint"              
                    async with self.session.post(url, data = f"[{json['packageFingerprint']}]") as response:
                        json = await response.json()
                        if json['exactMatches']:
                            return await self._get_curseforge(json['exactMatches'][0], hash)

        return None

    async def get_author(self, ID: str | int, provider: str) -> str:

        if provider == "Modrinth":

            async with self.session.get(f"{self.modrinth}/mod/{ID}") as response:

                json = await response.json()
                teamID = json['team']

                async with self.session.get(f"{self.modrinth}/team/{teamID}/members") as response:

                    json = await response.json()

                    for team_member in json:
                        if team_member['role'] == "Owner":
                            userID = team_member['user']['id']
                            break

                    async with self.session.get(f"{self.modrinth}/user/{userID}") as response:

                        json = await response.json()

                        author = json['username']

        elif provider == "CurseForge":

            async with self.session.get(f"{self.curseforge}/addon/{ID}") as response:

                json = await response.json()

                author = json['authors'][0]['name']

        return author

    async def _get_modrinth(self, json: dict[str]) -> Resource:

        ID = json['project_id']
        fileID = json['id']

        filename = json['files'][0]['filename']
        url = json['files'][0]['url']

        hashes = json['files'][0]['hashes']

        async with self.session.get(f"{self.modrinth}/project/{ID}") as response:

            json = await response.json()

            name = json['title']
            teamID = json['team']
            slug = json['slug'] if 'slug' in json else name

            client = json['client_side']
            server = json['server_side']
            summary = "both"

            if client == "required": summary = "client"
            elif server == "required": summary = "server"

            async with self.session.get(f"{self.modrinth}/team/{teamID}/members") as response:

                json = await response.json()

                for team_member in json:
                    if team_member['role'] == "Owner":
                        userID = team_member['user']['id']
                        break

                async with self.session.get(f"{self.modrinth}/user/{userID}") as response:

                    json = await response.json()

                    author = json['username']

                    resource = Resource (
                        name = name,
                        filename=filename,
                        side=Side(client, server, summary).to_dict(),
                        hashes=hashes,
                        downloads={"Modrinth": Provider(ID, fileID, url, slug, author).to_dict()}
                    )
                    
                    return resource

    async def _get_curseforge(self, json: dict[str], hash: str | int) -> Resource:
  
        ID = json['id']
        fileID = json['file']['id']

        filename = json['file']['fileName']
        url = json['file']['downloadUrl']

        async with self.session.get(f"{self.curseforge}/addon/{ID}") as response:

            json = await response.json()

            name = json['name']
            author = json['authors'][0]['name']
            slug = json['slug'] if 'slug' in json else name

            resource = Resource (
                name = name,
                filename=filename,
                side=Side("optional", "optional", "both").to_dict(),
                hashes={"murmur2": hash},
                downloads={"CurseForge": Provider(ID, fileID, url, slug, author).to_dict()}
            )

            return resource



import asyncio
from os import walk
from pathlib import Path
from aiohttp import ClientSession

from mc_converter.Helpers.resourceAPI import Resource
from mc_converter.Helpers.utils import get_hash
from .Helpers.abstractions import ModpackManager

from json import load as parse_json

class CurseForgeManager(ModpackManager):

    def __init__(self, path: Path | str, session: ClientSession, config: dict[str]) -> None:

        self.manifest = dict()
        self.modlist = list()

        super().__init__(path, session, config)

    async def get_resource(self, resource: dict[str]) -> None:
        
        ID = resource['projectID']
        fileID = resource['fileID']

        resource: Resource = await self.resource_manager.get_by_ID(ID, fileID)
        if resource is None: return

        data = {       
            "name": resource.name,
            "filename": resource.filename,

            "hashes": resource.hashes,

            "relative_path": "mods" if Path(resource.filename).suffix == ".jar" else "resourcepacks", # I don't know how to categorize it anyway

            "side": resource.side,
            "downloads": resource.downloads
        }

        self.config['resources'].append(data)

    def get_override(self, path: Path) -> None:

        overrides_dir = self.temp_dir / "overrides"
        relative_path = path.relative_to(overrides_dir)

        data = {
            "filename": path.name,

            "hashes": {
                "sha256": get_hash(path)
            },

            "full_path": path.as_posix(),
            "relative_path": relative_path.as_posix(),
        }
        
        self.config['overrides'].append(data)

    async def parse(self) -> None:

        from shutil import unpack_archive
        unpack_archive(self.modpack_path, self.temp_dir)

        with open(self.temp_dir / "manifest.json") as file:
            self.manifest = parse_json(file)

        for entry in ("author", "name", "version"):
            if entry in self.manifest and not self.config[entry]:
                self.config[entry] = self.manifest[entry]

        self.config['minecraft'] = self.manifest['minecraft']['version']

        for loader in self.manifest['minecraft']['modloaders']:

            if "fabric" in loader['id']:
                self.config['modloader']['type'] = "fabric"
            if "forge" in loader['id']:
                self.config['modloader']['type'] = "forge"

            self.config['modloader']['version'] = loader['id'].split("-")[-1]

        futures = []

        for resource in self.manifest['files']:
            futures.append(self.get_resource(resource))

        await asyncio.gather(*futures)

        overrides = [Path(file) for _, _, filenames in walk(self.temp_dir / "overrides") for file in filenames]

        for override in overrides:
            self.get_override(override)

    def write_manifest(self) -> None:

        self.manifest = {
            "minecraft": {
                "version": self.config['minecraft'],
                "modLoaders": [
                    {
                    "id": f"{self.config['modloader']['type']}-{self.config['modloader']['version']}",
                    "primary": True
                    }
                ]
            },
            "manifestType": "minecraftModpack",
            "manifestVersion": 1,
            "name": self.config['name'],
            "version": self.config['version'],
            "author": self.config['author'],

            "files": [],
            "overrides": "overrides"
        }

    def add_resource(self, resource: dict[str]) -> None:

        provider_data = resource['downloads']['CurseForge']

        data = {
            "projectID": provider_data['ID'],
            "fileID": provider_data['fileID'],
            "required": True
        }

        self.manifest['files'].append(data)
        mod_page_url = "https://www.curseforge.com/minecraft/mc-mods/" + provider_data['slug']
        self.modlist.append(f"<li><a href=\"{mod_page_url}\">{resource['name']} (by {provider_data['author']})</a></li>\n")

    def add_override(self, file: dict[str]) -> None:

        if "full_path" not in file: return

        overrides_dir = self.temp_dir / "overrides"
        overrides_dir.mkdir(parents=True, exist_ok=True)

        file_path = overrides_dir / file['relative_path']
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file['full_path'], file_path)

    def write(self) -> None:

        self.write_manifest()

        for resource in self.config['resources']:
            if "CurseForge" in resource['downloads']:
                self.add_resource(resource)
            else: self.add_override(resource)

        for override in self.config['overrides']:
            self.add_override(override)

        from json import dump as write_json
        with open(self.temp_dir / "manifest.json", 'w') as file:
            write_json(self.manifest, file)

        with open(self.temp_dir / "modlist.html", 'w') as file:
            self.modlist.insert(0, "<ul>\n")
            self.modlist.append("</ul>\n")
            file.writelines(self.modlist)

        from shutil import make_archive
        make_archive(self.modpack_path / ("CF_" + self.config['name']), 'zip', self.temp_dir, '.')



from os import walk
from pathlib import Path
from aiohttp import ClientSession

from json import load as parse_json

import asyncio

from .Helpers.abstractions import ModpackManager
from .Helpers.resourceAPI import Resource
from .Helpers.utils import get_hash

class ModrinthManager(ModpackManager):

    def __init__(self, path: Path | str, session: ClientSession, config: dict[str]) -> None:

        self.index = dict()

        super().__init__(path, session, config)

    async def get_resource(self, file: dict[str]) -> None:

        resource: Resource = None

        for k, v in file['hashes']:
            temp_resource = await self.resource_manager.get_by_hash(k, v)

            if resource is not None:
                resource.downloads.update(temp_resource.downloads)
                resource.hashes.update(temp_resource.hashes)
                continue

            resource = temp_resource

        if resource is None: return

        path = Path(file['path'])

        data = {
                
                "name": resource.name,
                "filename": resource.filename,

                "hashes": resource.hashes,

                "relative_path": path.parent.name,

                "side": resource.side,
                "downloads": resource.downloads
            }

        self.config['resources'].append(data)  

    def get_override(self, path: Path) -> None:

        overrides_dir = self.temp_dir / "overrides"
        relative_path = path.relative_to(overrides_dir)

        data = {
            "filename": path.name,

            "hashes": {
                "sha256": get_hash(path)
            },

            "full_path": path.as_posix(),
            "relative_path": relative_path.as_posix(),
        }
        
        self.config['overrides'].append(data)

    async def parse(self) -> None:

        from shutil import unpack_archive
        unpack_archive(self.modpack_path, self.temp_dir)

        with open(self.temp_dir / "modrinth.index.json") as file:
            self.index = parse_json(file)

        self.config['name'] = self.index['name']
        self.config['version'] = self.index['versionId']
        if "summary" in self.index:
            self.config['description'] = self.index['summary']

        self.config['minecraft'] = self.index['dependencies']['minecraft']

        match self.index['dependencies']:
            case {'fabric-loader': fabric_version}: 
                self.config['modloader']['type'] = "fabric"
                self.config['modloader']['version'] = fabric_version
            case {'forge': forge_version}: 
                self.config['modloader']['type'] = "forge"
                self.config['modloader']['version'] = forge_version

        futures = []

        for resource in self.index['files']:
            futures.append(self.get_resource(resource))

        await asyncio.gather(*futures)

        overrides = [Path(file) for _, _, filenames in walk(self.temp_dir / "overrides") for file in filenames]

        for override in overrides:
            self.get_override(override)

    def write_index(self): 

        match self.config['modloader']['type']:

            case "fabric": modloader = "fabric-loader"
            case _: modloader = self.config['modloader']['type']

        if self.config['modloader']['type'] == "fabric":
            modloader = "fabric-loader"
        
        self.index = {
            "formatVersion": 1,
            "game": "minecraft",

            "versionId": self.config['version'],
            "name": self.config['name'],
            "summary": self.config['description'],

            "files": [],

            "dependencies": {
                "minecraft": self.config['minecraft'],
                modloader: self.config['modloader']['version']
            }
        }

    def add_resource(self, resource: dict[str]) -> None:
        
        relative_path: Path = Path(resource['relative_path']) / resource['filename']

        data = {
            "path": relative_path.as_posix(),
            "hashes": resource['hashes'],
            "downloads": [resource['downloads'][provider]['url'] for provider in resource['downloads']]
        }

        self.index['files'].append(data)

    def add_override(self, file: dict[str]) -> None:

        overrides_dir = self.temp_dir / "overrides"
        overrides_dir.mkdir(parents=True, exist_ok=True)

        file_path = overrides_dir / file['relative_path']
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file['full_path'], file_path)

    def write(self) -> None:

        self.write_index()

        for override in self.config['overrides']:
            self.add_override(override)

        for resource in self.config['resources']:
            self.add_resource(resource)

        from json import dump as write_json
        with open(self.temp_dir / "modrinth.index.json", 'w') as file:
            write_json(self.index, file)

        from shutil import make_archive
        archive_name = make_archive(self.modpack_path / ("MR_" + self.config['name']), 'zip', self.temp_dir, '.')
        Path(archive_name).rename(archive_name.replace("zip", "mrpack"))



import asyncio
from pathlib import Path

from .Helpers.abstractions import ModpackManager
from .Helpers.utils import get_hash

class MultiMCManager(ModpackManager):

    async def get_resource(self, path: Path) -> None:

        data = {}

        if resource := await self.resource_manager.get(path):

            data = {
                
                "name": resource.name,
                "filename": resource.filename,

                "hashes": resource.hashes,

                "relative_path": path.parent.name,
                "full_path": path.as_posix(),

                "side": resource.side,
                "downloads": resource.downloads
            }

        elif "Resource" in self.config and path.name in [item['filename'] for item in self.config['Resource']]:

            resource = [item for item in self.config['Resource'] if item['filename'] == path.name][0]

            data = {

                "name": resource['name'],
                "filename": path.name,

                "hashes": {
                    "sha256": get_hash(path)
                },
                
                "relative_path": path.parent.name,
                "full_path": path.as_posix(),

                "side": {
                    "client": "optional",
                    "server": "optional",
                    "summary": "both"
                },

                "downloads": {
                    "Other": {
                        "url": resource['url'],
                        "slug": resource['slug'],
                        "author": "Unknown"
                    }
                }
            }

            self.config['Resource'].remove(resource)

        else: 
            print(f"File {path.name} not found of CF or MR")
            return self.get_override(path)

        self.config['resources'].append(data)

    def get_override(self, path: Path) -> None: 

        root_dir_id = path.parts.index(".minecraft")
        relative_path = path.relative_to(*path.parts[:root_dir_id + 1]).parent

        data = {
            "filename": path.name,

            "hashes": {
                "sha256": get_hash(path)
            },

            "full_path": path.as_posix(),
            "relative_path": relative_path.as_posix(),
        }
        
        self.config['overrides'].append(data)

    async def parse(self) -> None:

        downloadable_content = ("resourcepacks", "shaderpacks", "mods")

        from shutil import unpack_archive
        from os import walk
        
        unpack_archive(self.modpack_path, self.temp_dir)

        futures = list()
        overrides = list()

        for folder_path, _, filenames in walk(self.temp_dir):
            if ".minecraft" not in folder_path: continue
            for filename in filenames:
                filepath = Path(folder_path) / filename
                if folder_path.endswith(downloadable_content): 
                    future = self.get_resource(filepath)
                    futures.append(future)
                else: overrides.append(filepath)

        await asyncio.gather(*futures)
        for override in overrides:
            self.get_override(override)

        modpack_dir = next(self.temp_dir.iterdir())
        if not self.config['name']: self.config['name'] = modpack_dir.name

        from json import load as parse_json

        with open(modpack_dir / "mmc-pack.json") as file:
            json = parse_json(file)

        for component in json['components']:
            if component['cachedName'] == "Minecraft":
                if not self.config['minecraft']: 
                    self.config['minecraft'] = component['cachedVersion']
            elif component['cachedName'] == "Fabric Loader":
                if not self.config['modloader']['type']: 
                    self.config['modloader']['type'] = "fabric"
                    self.config['modloader']['version'] = component['cachedVersion']
            elif component['cachedName'] == "Forge":
                if not self.config['modloader']['type']: 
                    self.config['modloader']['type'] = "forge"
                    self.config['modloader']['version'] = component['cachedVersion']


    def write(self) -> None:
        return super().write()



from argparse import ArgumentParser
from aiohttp import ClientSession, TCPConnector
from pathlib import Path

from json import dump as write_json
from tomli import load as parse_toml

from .Helpers.utils import get_pack_format, get_default_config, get_pack_manager

async def program():

    formats = ['packwiz', 'modrinth', 'curseforge', 'intermediate']
 
    parser = ArgumentParser(description="Converts modpack formats to each other", exit_on_error=True)
    parser.add_argument('-c', '--config', dest='config', type=Path, help='Path to config, used to fill the gaps in parsed data')
    parser.add_argument('-i', '--input', dest='input', type=Path, help='Path to pack', required=True)
    parser.add_argument('-f', '--format', dest='formats', type=str, nargs="+", choices=formats, help='Format to convert to', required=True)
    parser.add_argument('-o', '--output', dest='output', type=Path, help='Specify output directory (optional)', default=Path.cwd())
    args = parser.parse_args()

    if not args.input.exists(): exit("Invalid input!")
    input_format = get_pack_format(args.input)
    if input_format == "Unknown": exit("Invalid pack input format!")

    config = get_default_config()

    if args.config:
        with open(args.config, "rb") as file:
            config.update(parse_toml(file))

    async with ClientSession(connector=TCPConnector(limit=0)) as session:
        input_manager = get_pack_manager(input_format)(args.input, session, config)
        await input_manager.parse()

        if not config['author']: config['author'] = input("Author: ")
        if not config['name']: config['name'] = input("Name: ")
        if not config['version']: config['version'] = input("Version: ")
        if not config['description']: config['description'] = input("Description: ")

        if "Resource" in config: del config['Resource']

        for format in args.formats:

            if format == "intermediate":
                with open(args.output / "intermediate_output.json", "w") as file:
                    write_json(config, file, indent=4)
                continue
            
            if input_format == format:
                print(f"{format} to {format}? Why? Anyway it's a bad idea. I will prevent it.")
                continue

            output_manager = get_pack_manager(format)(args.output, session, config)
            output_manager.write()

    return 0

def main():
    import sys
    import asyncio

    if sys.platform.startswith("win"):
        asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())
                
    loop = asyncio.get_event_loop()
    loop.run_until_complete(program())


from ctypes import ArgumentError
from os import walk
from pathlib import Path

def get_hash(path: Path, type: str = "sha256") -> str:
        
    from hashlib import sha1, sha256, sha512
    from murmurhash2 import murmurhash2 as murmur2

    with open(path, "rb") as file:
        data = file.read()

    match(type):

        case "sha1": hash = sha1(data).hexdigest()
        case "sha256": hash = sha256(data).hexdigest()
        case "sha512": hash = sha512(data).hexdigest()

        case "murmur2": 
            data = bytes([b for b in data if b not in (9, 10, 13, 32)])
            hash = murmur2(data, seed=1)

        case _: raise(ArgumentError("Incorrect hash type!"))

    return str(hash)

def get_pack_format(path: Path) -> str:

    # Key modpack identifiers:
    # Modrinth has index.json
    # MultiMC has mmc-pack.json and instance.cfg
    # CurseForge has manifest.json and modlist.html
    # packwiz can be a folder, and has index.toml and pack.toml
    # So.. That was easier then I thought

    if path.is_dir():
        files = [file for _, _, filenames in walk(path) for file in filenames]
        if "index.toml" in files and "pack.toml" in files: return "packwiz"

    if str(path).endswith("mrpack"):
        from zipfile import ZipFile
        with ZipFile(path) as zip:

            filenames = [Path(file).name for file in zip.namelist()]

            if "modrinth.index.json" in filenames: return "modrinth"

    if str(path).endswith("zip"):
        from zipfile import ZipFile
        with ZipFile(path) as zip:

            filenames = [Path(file).name for file in zip.namelist()]

            if "modrinth.index.json" in filenames: return "modrinth"
            elif "index.toml" in filenames and "pack.toml" in filenames: return "packwiz"
            elif "mmc-pack.json" in filenames and "instance.cfg" in filenames: return "multimc"
            elif "manifest.json" in filenames and "modlist.html" in filenames: return "curseforge"

    return "Unknown"

def get_default_config() -> dict[str]:

    config = {
                "author": str(),
                "name": str(),
                "version": str(),
                "description": str(),

                "minecraft": str(),
                "modloader": {
                    "type": str(),
                    "version": str()
                },

                "resources": [],
                "overrides": []
            }

    return config

from .abstractions import ModpackManager
from mc_converter.MultiMC import MultiMCManager
from mc_converter.CurseForge import CurseForgeManager
from mc_converter.Modrinth import ModrinthManager
from mc_converter.packwiz import packwizManager

def get_pack_manager(pack_format: str) -> ModpackManager:

    match pack_format:

        case "multimc": return MultiMCManager
        case "curseforge": return CurseForgeManager
        case "modrinth": return ModrinthManager
        case "packwiz": return packwizManager



from abc import ABC, abstractmethod
from pathlib import Path
from aiohttp import ClientSession

from .resourceAPI import ResourceAPI

class ModpackManager(ABC):

    def __init__(self, path: Path | str, session: ClientSession, config: dict[str]) -> None:

        from tempfile import TemporaryDirectory
        self._temp_dir = TemporaryDirectory()
        self.temp_dir = Path(self._temp_dir.name)

        self.modpack_path = Path(path)
        self.session = session
        self.config = config

        self.resource_manager = ResourceAPI(session)

    def __del__(self) -> None:
        self._temp_dir.cleanup()

    async def parse(self) -> None: 
        raise NotImplementedError()

    @abstractmethod
    def write(self) -> None:
        raise NotImplementedError()



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['mc_converter', 'mc_converter.Helpers']

package_data = \
{'': ['*']}

install_requires = \
['aiohttp>=3.7.4,<4.0.0',
 'murmurhash2>=0.2.9,<0.3.0',
 'pytoml>=0.1.21,<0.2.0',
 'tenacity>=8.0.1,<9.0.0',
 'tomli>=1.2.1,<2.0.0']

entry_points = \
{'console_scripts': ['mc-converter = mc_converter:main']}

setup_kwargs = {
    'name': 'mc-converter',
    'version': '1.0.3.4',
    'description': '',
    'long_description': "# modpack-converter\nConvert's one minecraft modpack format to another.\n\nMaybe I need to think of a better name...\n\n# Features\n\n- Auto detect input modpack format\n- Support convertions to and from:\n    - MultiMC (parsing only)\n    - CurseForge\n    - Modrinth\n    - packwiz\n- Detects downloadable resourcepacks and shaders (MultiMC only)\n- User friendly toml config\n- Multiple output formats at once\n\n# How to Use\n\n```\nmc-converter [-h] [-c CONFIG] -i INPUT -f FORMAT [-o OUTPUT]\n```\n\n## Explanation:\n\n```\n-h --help: prints help\n-i --input: specifies input file (mostly zip file)\n-c --config: specifies config file, used for fill the gaps like description or files not in modrinth on curseforge example can be found in this repository.\n-f --format: soecifies formats to convert, must be separated by space.\n-o --output: specifies output directory, where converted zip files will be stored. By default current working directory will be used.\n```\n\nAvaliable formats:     - `curseforge, modrinth, packwiz, intermediate`\n\n`intermediate` must be used only for debuging, can contain sensetive information\n\nExample: \n```\nmc-converter -i MyLovelyMultiMcModpack.zip -c config.toml -f curseforge modrinth -o converted_modpacks\n```\n\n# How to Install\n\n## From PyPI\n```\npip install mc-converter\n```\n## From `.whl` file\nGo to Release page, download latest `.whl` file \\\n And install it via the following comand:\n ```\n pip install mc_converter_{version}.whl\n ```\n \n # Credits\n\nmurmurhash2 - Murmur Hash 2 libray - https://pypi.org/project/murmurhash2 \\\naiohttp - Async web interface - https://github.com/aio-libs/aiohttp \\\ntomli - Fast pure python toml parser - https://github.com/hukkin/tomli \\\npytoml - The only one toml writer that can handle weird packwiz files - https://github.com/avakar/pytoml \\\ntenacity - Awesome, ease-in-use retrying library - https://github.com/jd/tenacity",
    'author': 'RozeFound',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/RozeFound/modpack-converter',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import asyncio
from pathlib import Path
from aiohttp.client import ClientSession

from tomli import load as parse_toml
from pytoml import dump as write_toml

from .Helpers.abstractions import ModpackManager
from .Helpers.utils import get_hash

class packwizManager(ModpackManager):

    def __init__(self, path: Path | str, session: ClientSession, config: dict[str]) -> None:

        self.pack_info = dict()

        self.index = {
            "hash-format": "sha256",
            "files": []
        }

        super().__init__(path, session, config)

    async def get_resource(self, resource: dict[str]) -> None:

        toml_path: Path = self.temp_dir / resource['file']
        with open(toml_path, "rb") as file:
            toml_data = parse_toml(file)

        data = {
            "name": toml_data["name"],
            "filename": toml_data['filename'],

            "hashes": {
                toml_data['download']['hash-format']: toml_data['download']['hash']
            },

            "relative_path": toml_path.parts[-2],

            "side": {
                "client": "required" if toml_data['side'] == "client" else "optional",
                "server": "required" if toml_data['side'] == "server" else "optional",
                "summary": toml_data['side']
            },

            "downloads": {}
        }

        if "update" in toml_data:

            match toml_data['update']:

                case {"curseforge": cf}:

                    data['downloads']['CurseForge'] = {
                        "ID": cf['project-id'],
                        "fileID": cf['file-id'],
                        "url": toml_data['download']['url'],
                        "slug": toml_path.stem,
                        "author": await self.resource_manager.get_author(cf['project-id'], "CurseForge")
                    }

                case {"modrinth": mr}:

                    data['downloads']['Modrinth'] = {
                        "ID": mr['mod-id'],
                        "fileID": mr['version'],
                        "url": toml_data['download']['url'],
                        "slug": toml_path.stem,
                        "author": await self.resource_manager.get_author(mr['mod-id'], "Modrinth")
                    }
        else: 
            data['downloads']['Other'] = {
                "url": toml_data['download']['url'],
                "slug": toml_path.stem,
                "author": "Unknown"
            }

        self.config['resources'].append(data)

    def get_override(self, file: dict[str]) -> None:

        relaitve_path = Path(file['file'])
        full_path = self.temp_dir / relaitve_path

        data = {
            "filename": relaitve_path.name,

            "hashes": {
                "sha256": file['hash']
            },

            "full_path": full_path.as_posix(),
            "relative_path": relaitve_path.parent.as_posix()
        }

        self.config['overrides'].append(data)
    
    async def parse(self) -> None:

        if self.modpack_path.is_dir():
            self.temp_dir = self.modpack_path
        else:
            from shutil import unpack_archive
            unpack_archive(self.modpack_path, self.temp_dir)

        with open(self.temp_dir / "pack.toml", "rb") as file:
            self.pack_info = parse_toml(file)

        with open(self.temp_dir / self.pack_info['index']['file'], "rb") as file:
            self.index = parse_toml(file)

        for info_type in ("author", "name", "version", "description"):
            if info_type in self.pack_info and not self.config[info_type]:
                self.config[info_type] = self.pack_info[info_type]

        self.config['minecraft'] = self.pack_info['versions']['minecraft']

        match self.pack_info['versions']:
                case {'fabric': fabric_version}: 
                    self.config['modloader']['type'] = "fabric"
                    self.config['modloader']['version'] = fabric_version
                case {'forge': forge_version}:
                    self.config['modloader']['type'] = "forge"
                    self.config['modloader']['version'] = forge_version

        futures = []

        for file in self.index['files']:
            if "metafile" in file and file['metafile'] == True:
                futures.append(self.get_resource(file))
            else: self.get_override(file)

        await asyncio.gather(*futures)

    def write_pack_info(self) -> None:
        
        self.pack_info = {
            "name": self.config['name'],
            "author": self.config['author'],
            "version": self.config['version'],

            "index": {
                "file": "index.toml",
                "hash-format": "sha256",
                "hash": "placeholder"
            },

            "versions": {
                self.config['modloader']['type']: self.config['modloader']['version'],
                "minecraft": self.config['minecraft']
            }
        }

    def add_resource(self, resource: dict[str]) -> None:

        data = {
            "name": resource['name'],
            "filename": resource['filename'],
            "side": resource['side']['summary']
        }

        for provider in ("Modrinth", "CurseForge", "Other"):
            if provider in resource['downloads']:
                provider_data = resource['downloads'][provider]

                slug = provider_data['slug']

                match provider:
                    case "Modrinth": 
                        hash_format = "sha512" if "sha512" in resource['hashes'] else "sha1"
                        data['update'] = {
                            "modrinth": {
                                "mod-id": provider_data['ID'],
                                "version": provider_data['fileID']
                            }
                        }
                    case "CurseForge":
                        hash_format = "murmur2"
                        data['update'] = {
                            "curseforge": {
                                "file-id": provider_data['fileID'],
                                "project-id": provider_data['ID'],
                                "release-channel": "beta"
                            } 
                        }
                    case "Other": hash_format = "sha256"

                data['download'] = {
                    "url": provider_data['url'],
                    "hash-format": hash_format,
                    "hash": resource['hashes'][hash_format]
                }

                break
        
        toml_path = self.temp_dir / resource['relative_path'] / (slug + ".toml")
        toml_path.parent.mkdir(parents=True, exist_ok=True)

        with open(toml_path, "w") as file:
            write_toml(data, file)

        index_data = {
            "file": toml_path.relative_to(self.temp_dir).as_posix(),
            "hash": get_hash(toml_path),
            "metafile": True
        }
        
        self.index['files'].append(index_data)

    def add_override(self, file: dict[str]) -> None:

        file_path = self.temp_dir / file['relative_path']
        file_path.mkdir(parents=True, exist_ok=True)

        from shutil import copy2 as copy_file
        copy_file(file['full_path'], file_path)

        data = {
            "path": file['relative_path'] + "/" + file['filename'],
            "hash": file['hashes']['sha256']
        }
        
        self.index['files'].append(data)

    def write(self) -> None:

        for override in self.config['overrides']:
            self.add_override(override)

        for resource in self.config['resources']:
            self.add_resource(resource)

        index_path = self.temp_dir / "index.toml"
        with open(index_path, "w") as file:
            write_toml(self.index, file)

        self.write_pack_info()
        self.pack_info['index']['hash'] = get_hash(index_path)

        with open(self.temp_dir / "pack.toml", "w") as file:
            write_toml(self.pack_info, file)

        from shutil import make_archive
        make_archive(self.modpack_path / ("PW_" + self.config['name']), 'zip', self.temp_dir, '.')
